from dataclasses import dataclass, field


@dataclass
class Template:
    name: str = ""
    path: str = ""
    extends: str | None = None
    blocks: list[str] | None = None
    context: dict = field(default_factory=dict)


@dataclass
class Url:
    name: str = ""
    docs: str = ""
    source: str = ""


@dataclass
class Tag:
    name: str = ""
    docs: str = ""
    source: str = ""
    inner_tags: list[str] = ""
    closing_tag: str = ""


@dataclass
class Filter:
    name: str = ""
    docs: str = ""
    source: str = ""


@dataclass
class Library:
    name: str = ""
    tags: dict[str, Tag] = field(default_factory=dict)
    filters: dict[str, Filter] = field(default_factory=dict)


@dataclass
class WorkspaceIndex:
    src_path: str = ""
    env_path: str = ""
    file_watcher_globs: [str] = field(default_factory=list)
    static_files: [str] = field(default_factory=list)
    urls: dict[str, Url] = field(default_factory=dict)
    libraries: dict[str, Library] = field(default_factory=dict)
    templates: dict[str, Template] = field(default_factory=dict)
    global_template_context: dict[str, str] = field(default_factory=dict)

    def update(self, django_data: dict):
        self.file_watcher_globs = django_data.get(
            "file_watcher_globs", self.file_watcher_globs
        )
        self.static_files = django_data.get("static_files", self.static_files)
        self.urls = {
            name: Url(
                name=name,
                docs=options.get("docs", ""),
                source=options.get("source", ""),
            )
            for name, options in django_data.get("urls", {}).items()
        }

        self.libraries = {
            lib_name: Library(
                name=lib_name,
                filters={
                    name: Filter(
                        name=name,
                        docs=filter_options.get("docs", ""),
                        source=filter_options.get("source", ""),
                    )
                    for name, filter_options in lib_data.get("filters", {}).items()
                },
                tags={
                    tag: Tag(
                        name=tag,
                        docs=tag_options.get("docs"),
                        source=tag_options.get("source", ""),
                        inner_tags=tag_options.get("inner_tags", []),
                        closing_tag=tag_options.get("closing_tag"),
                    )
                    for tag, tag_options in lib_data.get("tags", {}).items()
                },
            )
            for lib_name, lib_data in django_data.get("libraries", {}).items()
        }

        self.templates = {
            name: Template(name=name, **options)
            for name, options in django_data.get("templates", {}).items()
        }

        self.global_template_context = django_data.get("global_template_context", {})



import logging
import re
from functools import cached_property
from re import Match
from textwrap import dedent

import jedi
from lsprotocol.types import CompletionItem, Hover, Location, Position, Range
from pygls.workspace import TextDocument

from djlsp.constants import BUILTIN
from djlsp.index import WorkspaceIndex

logger = logging.getLogger(__name__)


class TemplateParser:

    def __init__(
        self,
        workspace_index: WorkspaceIndex,
        jedi_project: jedi.Project,
        document: TextDocument,
    ):
        self.workspace_index: WorkspaceIndex = workspace_index
        self.jedi_project: jedi.Project = jedi_project
        self.document: TextDocument = document

    @cached_property
    def loaded_libraries(self):
        re_loaded = re.compile(r".*{% ?load ([\w ]*) ?%}$")
        loaded = {BUILTIN}
        for line in self.document.lines:
            if match := re_loaded.match(line):
                loaded.update(
                    [
                        lib
                        for lib in match.group(1).strip().split(" ")
                        if lib in self.workspace_index.libraries
                    ]
                )
        logger.debug(f"Loaded libraries: {loaded}")
        return loaded

    @cached_property
    def context(self):
        context = self.workspace_index.global_template_context.copy()
        if "/templates/" in self.document.path:
            template_name = self.document.path.split("/templates/", 1)[1]
            if template := self.workspace_index.templates.get(template_name):
                context.update(template.context)

        # Add all variables found in template to context
        # TODO: Use scope to only add to context based on cursor position
        re_as = re.compile(r".*{%.*as ([\w ]+) %}.*$")
        re_for = re.compile(r".*{% ?for ([\w ,]*) in.*$")
        re_with = re.compile(r".*{% ?with (.+) ?%}.*")
        found_variables = []
        for line in self.document.lines:
            if match := re_as.match(line):
                found_variables.extend(match.group(1).split(" "))
            if match := re_for.match(line):
                context["forloop"] = None
                found_variables.extend(match.group(1).split(","))
            if match := re_with.match(line):
                for assignment in match.group(1).split(" "):
                    split_assignment = assignment.split("=")
                    if len(split_assignment) == 2:
                        found_variables.append(split_assignment[0])

        for variable in found_variables:
            if variable_stripped := variable.strip():
                context[variable_stripped] = None

        # Update type definations based on template type comments
        # only simple version of variable: full python path:
        # {# type some_variable: full.python.path.to.class #}
        re_type = re.compile(r".*{# type (\w+) ?: ?([\w\d_\.]+) ?#}.*")
        for line in self.document.lines:
            if match := re_type.match(line):
                variable = match.group(1)
                variable_type = match.group(2)
                context[variable] = variable_type

        return context

    def create_jedi_script(self, code) -> jedi.Script:
        """
        Generate jedi Script based on template context and given code.
        """
        script_lines = []
        if re.search(r"{% ?for ", self.document.source):
            # TODO: Only add in for scope
            script_lines.append(
                dedent(
                    """
                    class DummyForLoop:
                        counter: int
                        counter0: int
                        revcounter: int
                        revcounter0: int
                        first: bool
                        last: bool
                        parentloop: "DummyForLoop"
                    forloop: DummyForLoop
                    """
                )
            )
        for variable, variable_type in self.context.items():
            if variable_type:
                variable_import = ".".join(variable_type.split(".")[:-1])
                script_lines.extend(
                    [
                        f"import {variable_import}",
                        f"{variable}: {variable_type}",
                    ]
                )
            else:
                script_lines.append(f"{variable} = None")

        # Add user code
        script_lines.append(code)

        return jedi.Script(code="\n".join(script_lines), project=self.jedi_project)

    ###################################################################################
    # Completions
    ###################################################################################
    def completions(self, line, character):
        line_fragment = self.document.lines[line][:character]
        matchers = [
            (re.compile(r".*{% ?load ([\w ]*)$"), self.get_load_completions),
            (re.compile(r".*{% ?block ([\w]*)$"), self.get_block_completions),
            (re.compile(r""".*{% ?url ('|")([\w\-:]*)$"""), self.get_url_completions),
            (
                re.compile(r".*{% ?static ('|\")([\w\-\.\/]*)$"),
                self.get_static_completions,
            ),
            (
                re.compile(r""".*{% ?(extends|include) ('|")([\w\-:]*)$"""),
                self.get_template_completions,
            ),
            (re.compile(r"^.*{% ?(\w*)$"), self.get_tag_completions),
            (
                re.compile(r"^.*({%|{{).*?\|(\w*)$"),
                self.get_filter_completions,
            ),
            (
                re.compile(r"^.*{# type \w+ ?: ?([\w\d_\.]*)$"),
                self.get_type_comment_complations,
            ),
            (
                re.compile(r".*({{|{% \w+ ).*?([\w\d_\.]*)$"),
                self.get_context_completions,
            ),
        ]

        for regex, completion in matchers:
            if match := regex.match(line_fragment):
                # Sort completions because some editors (Helix) will use order
                # as is and wont use sort_text.
                return list(
                    sorted(
                        completion(match),
                        key=lambda comp: (
                            comp.sort_text if comp.sort_text else comp.label
                        ),
                    )
                )
        return []

    def get_load_completions(self, match: Match):
        prefix = match.group(1).split(" ")[-1]
        logger.debug(f"Find load matches for: {prefix}")
        return [
            CompletionItem(label=lib)
            for lib in self.workspace_index.libraries.keys()
            if lib != BUILTIN and lib.startswith(prefix)
        ]

    def get_block_completions(self, match: Match):
        prefix = match.group(1).strip()
        logger.debug(f"Find block matches for: {prefix}")
        block_names = []
        if "/templates/" in self.document.path:
            template_name = self.document.path.split("/templates/", 1)[1]
            if template := self.workspace_index.templates.get(template_name):
                block_names = self._recursive_block_names(template.extends)

        used_block_names = []
        re_block = re.compile(r"{% ?block ([\w]*) ?%}")
        for line in self.document.lines:
            if matches := re_block.findall(line):
                used_block_names.extend(matches)

        return [
            CompletionItem(label=name)
            for name in block_names
            if name not in used_block_names and name.startswith(prefix)
        ]

    def _recursive_block_names(self, template_name, looked_up_templates=None):
        looked_up_templates = looked_up_templates if looked_up_templates else []
        looked_up_templates.append(template_name)

        block_names = []
        if template := self.workspace_index.templates.get(template_name):
            block_names.extend(template.blocks)
            if template.extends and template.extends not in looked_up_templates:
                block_names.extend(
                    self._recursive_block_names(template.extends, looked_up_templates)
                )
        return list(set(block_names))

    def get_static_completions(self, match: Match):
        prefix = match.group(2)
        logger.debug(f"Find static matches for: {prefix}")
        return [
            CompletionItem(label=static_file)
            for static_file in self.workspace_index.static_files
            if static_file.startswith(prefix)
        ]

    def get_url_completions(self, match: Match):
        prefix = match.group(2)
        logger.debug(f"Find url matches for: {prefix}")
        return [
            CompletionItem(label=url.name, documentation=url.docs)
            for url in self.workspace_index.urls.values()
            if url.name.startswith(prefix)
        ]

    def get_template_completions(self, match: Match):
        prefix = match.group(3)
        logger.debug(f"Find {match.group(1)} matches for: {prefix}")
        return [
            CompletionItem(label=template)
            for template in self.workspace_index.templates
            if template.startswith(prefix)
        ]

    def get_tag_completions(self, match: Match):
        prefix = match.group(1)
        logger.debug(f"Find tag matches for: {prefix}")

        tags = []
        for lib_name in self.loaded_libraries:
            if lib := self.workspace_index.libraries.get(lib_name):
                for tag in lib.tags.values():
                    tags.append(
                        CompletionItem(
                            label=tag.name,
                            documentation=tag.docs,
                        )
                    )
                    # TODO: Only add inner/clossing if there is opening tag
                    tags.extend([CompletionItem(label=tag) for tag in tag.inner_tags])
                    if tag.closing_tag:
                        tags.append(CompletionItem(label=tag.closing_tag))

        return [tag for tag in tags if tag.label.startswith(prefix)]

    def get_filter_completions(self, match: Match):
        prefix = match.group(2)
        logger.debug(f"Find filter matches for: {prefix}")
        filters = []
        for lib_name in self.loaded_libraries:
            if lib := self.workspace_index.libraries.get(lib_name):
                filters.extend(
                    [
                        CompletionItem(
                            label=filt.name,
                            documentation=filt.docs,
                        )
                        for filt in lib.filters.values()
                    ]
                )
        return [
            filter_name
            for filter_name in filters
            if filter_name.label.startswith(prefix)
        ]

    def get_type_comment_complations(self, match: Match):
        prefix = match.group(1)
        logger.debug(f"Find type comment matches for: {prefix}")

        if "." in prefix:
            from_part = ".".join(prefix.split(".")[:-1])
            import_part = prefix.split(".")[-1]
            code = f"from {from_part} import {import_part}"
        else:
            code = f"import {prefix}"

        return [
            CompletionItem(label=comp.name)
            for comp in self.create_jedi_script(code).complete()
        ]

    def get_context_completions(self, match: Match):
        prefix = match.group(2)
        logger.debug(f"Find context matches for: {prefix}")

        def get_sort_text(comp):
            type_sort = {"statement": "1", "property": "2"}.get(comp.type, "9")
            return f"{type_sort}-{comp.name}".lower()

        if "." in prefix:
            # Find . completions with Jedi
            return [
                CompletionItem(label=comp.name, sort_text=get_sort_text(comp))
                for comp in self.create_jedi_script(prefix).complete()
                if not comp.name.startswith("_")
            ]
        else:
            # Only context completions
            return [
                CompletionItem(label=var, sort_text=var.lower())
                for var in self.context
                if var.startswith(prefix)
            ]

    ###################################################################################
    # Hover
    ###################################################################################
    def hover(self, line, character):
        line_fragment = self.document.lines[line][:character]
        matchers = [
            (re.compile(r""".*{% ?url ('|")([\w\-:]*)$"""), self.get_url_hover),
            (re.compile(r"^.*({%|{{) ?[\w \.\|]*\|(\w*)$"), self.get_filter_hover),
            (re.compile(r"^.*{% ?(\w*)$"), self.get_tag_hover),
        ]
        for regex, hover in matchers:
            if match := regex.match(line_fragment):
                return hover(line, character, match)
        return None

    def get_url_hover(self, line, character, match: Match):
        full_match = self._get_full_hover_name(
            line, character, match.group(2), regex=r"^([\w\d:\-]+).*"
        )
        logger.debug(f"Find url hover for: {full_match}")
        if url := self.workspace_index.urls.get(full_match):
            return Hover(
                contents=url.docs,
            )

    def get_filter_hover(self, line, character, match: Match):
        filter_name = self._get_full_hover_name(line, character, match.group(2))
        logger.debug(f"Find filter hover for: {filter_name}")
        for lib in self.workspace_index.libraries.values():
            if lib.name in self.loaded_libraries and filter_name in lib.filters:
                return Hover(
                    contents=lib.filters[filter_name].docs,
                )
        return None

    def get_tag_hover(self, line, character, match: Match):
        tag_name = self._get_full_hover_name(line, character, match.group(1))
        logger.debug(f"Find tag hover for: {tag_name}")
        for lib in self.workspace_index.libraries.values():
            if lib.name in self.loaded_libraries and tag_name in lib.tags:
                return Hover(
                    contents=lib.tags[tag_name].docs,
                )
        return None

    def _get_full_hover_name(self, line, character, first_part, regex=r"^([\w\d]+).*"):
        if match_after := re.match(regex, self.document.lines[line][character:]):
            return first_part + match_after.group(1)
        return first_part

    ###################################################################################
    # Goto definition
    ###################################################################################
    def goto_definition(self, line, character):
        line_fragment = self.document.lines[line][:character]
        matchers = [
            (
                re.compile(r""".*{% ?(extends|include) ('|")([\w\-\./]*)$"""),
                self.get_template_definition,
            ),
            (re.compile(r""".*{% ?url ('|")([\w\-:]*)$"""), self.get_url_definition),
            (re.compile(r"^.*{% ?(\w*)$"), self.get_tag_definition),
            (re.compile(r"^.*({%|{{).*?\|(\w*)$"), self.get_filter_definition),
            (
                re.compile(r".*({{|{% \w+ ).*?([\w\d_\.]*)$"),
                self.get_context_definition,
            ),
        ]
        for regex, definition in matchers:
            if match := regex.match(line_fragment):
                return definition(line, character, match)
        return None

    def create_location(self, location, path, line):
        root_path = (
            self.workspace_index.src_path
            if location == "src"
            else self.workspace_index.env_path
        )
        return Location(
            uri=f"file://{root_path}/{path}",
            range=Range(
                start=Position(line=int(line), character=0),
                end=Position(line=int(line), character=0),
            ),
        )

    def get_template_definition(self, line, character, match: Match):
        if match_after := re.match(r'^(.*)".*', self.document.lines[line][character:]):
            template_name = match.group(3) + match_after.group(1)
            logger.debug(f"Find template goto definition for: {template_name}")
            if template := self.workspace_index.templates.get(template_name):
                location, path = template.path.split(":")
                return self.create_location(location, path, 0)

    def get_url_definition(self, line, character, match: Match):
        full_match = self._get_full_definition_name(
            line, character, match.group(2), regex=r"^([\w\d:\-]+).*"
        )
        logger.debug(f"Find url goto definition for: {full_match}")
        if url := self.workspace_index.urls.get(full_match):
            if url.source:
                return self.create_location(*url.source.split(":"))

    def get_tag_definition(self, line, character, match: Match):
        full_match = self._get_full_definition_name(line, character, match.group(1))
        logger.debug(f"Find tag goto definition for: {full_match}")
        for lib in self.loaded_libraries:
            if tag := self.workspace_index.libraries[lib].tags.get(full_match):
                if tag.source:
                    return self.create_location(*tag.source.split(":"))

    def get_filter_definition(self, line, character, match: Match):
        full_match = self._get_full_definition_name(line, character, match.group(2))
        logger.debug(f"Find filter goto definition for: {full_match}")
        for lib in self.loaded_libraries:
            if filter_ := self.workspace_index.libraries[lib].filters.get(full_match):
                if filter_.source:
                    return self.create_location(*filter_.source.split(":"))

    def get_context_definition(self, line, character, match: Match):
        first_match = match.group(2)
        full_match = self._get_full_definition_name(line, character, first_match)
        logger.debug(f"Find context goto definition for: {full_match}")
        if gotos := self.create_jedi_script(full_match).goto(column=len(first_match)):
            goto = gotos[0]
            if goto.module_name == "__main__":
                # Location is in fake script get type location
                if infers := goto.infer():
                    goto = infers[0]
                else:
                    return None
            return Location(
                uri=f"file://{goto.module_path}",
                range=Range(
                    start=Position(line=goto.line, character=goto.column),
                    end=Position(line=goto.line, character=goto.column),
                ),
            )

    def _get_full_definition_name(
        self, line, character, first_part, regex=r"^([\w\d]+).*"
    ):
        if match_after := re.match(regex, self.document.lines[line][character:]):
            return first_part + match_after.group(1)
        return first_part



import json
import os
import shutil
import subprocess

from djlsp.index import WorkspaceIndex
from djlsp.server import DJANGO_COLLECTOR_SCRIPT_PATH

DJANGO_TEST_PROJECT_SRC = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    "django_test",
)
DJANGO_TEST_SETTINGS_MODULE = "django_test.settings"


def test_django_collect():
    django_data = json.loads(
        subprocess.check_output(
            [
                shutil.which("python"),  # Get tox python path
                DJANGO_COLLECTOR_SCRIPT_PATH,
                f"--django-settings-module={DJANGO_TEST_SETTINGS_MODULE}",
                f"--project-src={DJANGO_TEST_PROJECT_SRC}",
            ]
        )
    )

    index = WorkspaceIndex()
    index.update(django_data)

    assert "django_app_tag" in index.libraries["django_app"].tags
    assert index.libraries["django_app"].tags["django_app_tag"].docs == "Docs for tag"

    assert "django_app_filter" in index.libraries["django_app"].filters
    assert (
        index.libraries["django_app"].filters["django_app_filter"].docs
        == "Docs for filter"
    )

    assert "django_app.html" in index.templates
    assert (
        index.templates["django_app.html"].path
        == "src:django_app/templates/django_app.html"
    )

    assert "django_app.js" in index.static_files
    assert "django_app:index" in index.urls
    assert set(index.file_watcher_globs) == {
        "**/templates/**",
        "**/templatetags/**",
        "**/static/**",
        "**/test-templates-folder/**",
        "**/test-static-folder/**",
    }



import argparse
import importlib
import inspect
import json
import logging
import os
import re
import sys
from dataclasses import dataclass
from unittest.mock import MagicMock, patch

import django
from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.staticfiles.finders import get_finders
from django.template.backends.django import get_installed_libraries
from django.template.engine import Engine
from django.template.library import InvalidTemplateLibrary
from django.template.utils import get_app_template_dirs
from django.urls import URLPattern, URLResolver
from django.views.generic import View
from django.views.generic.detail import SingleObjectMixin
from django.views.generic.edit import FormMixin
from django.views.generic.list import MultipleObjectMixin

logger = logging.getLogger(__name__)

# Some tags are added with a Node, like end*, elif else.
# TODO: Find a way of collecting these, for now hardcoded list
LIBRARIES_NODE_TAGS = {
    "__builtins__": {
        "autoescape": {
            "closing_tag": "endautoescape",
        },
        "filter": {
            "closing_tag": "endfilter",
        },
        "for": {
            "inner_tags": [
                "empty",
            ],
            "closing_tag": "endfor",
        },
        "if": {
            "inner_tags": [
                "else",
                "elif",
            ],
            "closing_tag": "endif",
        },
        "ifchanged": {
            "closing_tag": "endifchanged",
        },
        "spaceless": {
            "closing_tag": "endspaceless",
        },
        "verbatim": {
            "closing_tag": "endverbatim",
        },
        "with": {
            "closing_tag": "endwith",
        },
        "block": {
            "closing_tag": "endblock",
        },
        "comment": {
            "closing_tag": "endcomment",
        },
    },
    "cache": {
        "cache": {
            "closing_tag": "endcache",
        }
    },
    "i18n": {
        "language": {
            "closing_tag": "endlanguage",
        },
        "blocktrans": {
            "inner_tags": ["plural"],
            "closing_tag": "endblocktrans",
        },
        "blocktranslate": {
            "inner_tags": ["plural"],
            "closing_tag": "endblocktranslate",
        },
    },
    "l10n": {
        "localize": {
            "closing_tag": "endlocalize",
        }
    },
    "tz": {
        "localtime": {
            "closing_tag": "endlocaltime",
        },
        "timezone": {
            "closing_tag": "endtimezone",
        },
    },
}

# Context processors are functions and therefore hard to parse
# Use hardcoded mapping for know context processors.
TEMPLATE_CONTEXT_PROCESSORS = {
    # Django
    "django.template.context_processors.csrf": {
        "csrf_token": None,
    },
    "django.template.context_processors.debug": {
        "debug": None,
        "sql_queries": None,
    },
    "django.template.context_processors.i18n": {
        "LANGUAGES": None,
        "LANGUAGE_CODE": None,
        "LANGUAGE_BIDI": None,
    },
    "django.template.context_processors.tz": {
        "TIME_ZONE": None,
    },
    "django.template.context_processors.static": {
        "STATIC_URL": None,
    },
    "django.template.context_processors.media": {
        "MEDIA_URL": None,
    },
    "django.template.context_processors.request": {
        "request": None,
    },
    # Django: auth
    "django.contrib.auth.context_processors.auth": {
        "user": None,
        "perms": None,
    },
    # Django: messages
    "django.contrib.messages.context_processors.messages": {
        "messages": None,
        "DEFAULT_MESSAGE_LEVELS": None,
    },
    # Wagtail: settings
    "wagtail.contrib.settings.context_processors.settings": {
        # TODO: add fake settings object type with reference to models
        "settings": None,
    },
}


#######################################################################################
# Index Types
#######################################################################################
@dataclass
class Template:
    path: str = ""
    name: str = ""
    content: str = ""


#######################################################################################
# Index collector
#######################################################################################
class DjangoIndexCollector:
    re_extends = re.compile(r""".*{% ?extends ['"](.*)['"] ?%}.*""")
    re_block = re.compile(r".*{% ?block (\w*) ?%}.*")

    def __init__(self, project_src_path):
        self.project_src_path = project_src_path

        # Index data
        self.file_watcher_globs = []
        self.static_files = []
        self.urls = {}
        self.libraries = {}
        self.templates: dict[str, Template] = {}
        self.global_template_context = {}

    def collect(self):
        self.file_watcher_globs = self.get_file_watcher_globs()
        self.static_files = self.get_static_files()
        self.templates = self.get_templates()
        self.urls = self.get_urls()
        self.libraries = self.get_libraries()
        self.global_template_context = self.get_global_template_context()

        # Third party collectors
        self.collect_for_wagtail()

    def to_json(self):
        return json.dumps(
            {
                "file_watcher_globs": self.file_watcher_globs,
                "static_files": self.static_files,
                "urls": self.urls,
                "libraries": self.libraries,
                "templates": self.templates,
                "global_template_context": self.global_template_context,
            },
            indent=4,
        )

    def get_source_from_type(self, type_):
        def unwrap(func):
            while hasattr(func, "__wrapped__"):
                func = func.__wrapped__
            return func

        type_ = unwrap(type_)

        try:
            source_file = inspect.getsourcefile(type_)
            line = inspect.getsourcelines(type_)[1]
        except Exception as e:
            logger.error(e)
            return ""

        if source_file.startswith(self.project_src_path):
            path = source_file.removeprefix(self.project_src_path).lstrip("/")
            return f"src:{path}:{line}"
        elif source_file.startswith(sys.prefix):
            path = source_file.removeprefix(sys.prefix).lstrip("/")
            return f"env:{path}:{line}"
        return ""

    def get_type_full_name(self, type_):
        return f"{type_.__module__}.{type_.__name__}"

    # File watcher globs
    # ---------------------------------------------------------------------------------
    def get_file_watcher_globs(self):
        """
        File watcher glob patterns used to trigger this collector script

        https://microsoft.github.io/language-server-protocol/specifications/lsp/3.17/specification/#pattern
        """
        patterns = [
            "**/templates/**",
            "**/templatetags/**",
            "**/static/**",
        ]

        for static_path in settings.STATICFILES_DIRS:
            static_folder = os.path.basename(static_path)
            if static_folder != "static":
                patterns.append(f"**/{static_folder}/**")

        for template_path in [
            *Engine.get_default().dirs,
            *get_app_template_dirs("templates"),
        ]:
            template_folder = os.path.basename(template_path)
            if template_folder != "templates":
                patterns.append(f"**/{template_folder}/**")

        return patterns

    # Static files
    # ---------------------------------------------------------------------------------
    def get_static_files(self):
        # TODO: Add option to ignore some static folders
        # (like static that is generated with a JS bundler)
        static_paths = []
        for finder in get_finders():
            for path, _ in finder.list(None):
                static_paths.append(path)
        return static_paths

    # Urls
    # ---------------------------------------------------------------------------------
    def get_urls(self):
        try:
            urlpatterns = __import__(settings.ROOT_URLCONF, {}, {}, [""]).urlpatterns
        except Exception:
            return {}

        def recursive_get_views(urlpatterns, namespace=None, pattern=""):
            views = {}
            for p in urlpatterns:
                if isinstance(p, URLPattern):
                    # TODO: Get view path/line and template context
                    if not p.name:
                        name = p.name
                    elif namespace:
                        name = "{0}:{1}".format(namespace, p.name)
                    else:
                        name = p.name

                    if name:
                        callback = getattr(p.callback, "view_class", p.callback)
                        try:
                            self.add_template_context_for_view(callback)
                        except Exception:
                            pass
                        views[name] = {
                            "docs": f"{pattern}{p.pattern}",
                            "source": self.get_source_from_type(callback),
                        }
                elif isinstance(p, URLResolver):
                    try:
                        patterns = p.url_patterns
                    except ImportError:
                        continue
                    if namespace and p.namespace:
                        _namespace = "{0}:{1}".format(namespace, p.namespace)
                    else:
                        _namespace = p.namespace or namespace
                    views.update(
                        recursive_get_views(
                            patterns,
                            namespace=_namespace,
                            pattern=f"{pattern}{p.pattern}",
                        )
                    )
            return views

        return recursive_get_views(urlpatterns)

    # Libaries
    # ---------------------------------------------------------------------------------
    def get_libraries(self):
        libraries = {
            "__builtins__": {
                "tags": {},
                "filters": {},
            }
        }

        # Collect builtins
        for lib_mod_path in Engine.get_default().builtins:
            lib = importlib.import_module(lib_mod_path).register
            parsed_lib = self._parse_library(lib)
            libraries["__builtins__"]["tags"].update(parsed_lib["tags"])
            libraries["__builtins__"]["filters"].update(parsed_lib["filters"])

        # Get Django templatetags
        django_path = inspect.getabsfile(django.templatetags)
        django_mod_files = os.listdir(os.path.dirname(django_path))
        for django_lib in [
            i[:-3] for i in django_mod_files if i.endswith(".py") and i[0] != "_"
        ]:
            try:
                lib = get_installed_libraries()[django_lib]
                lib = importlib.import_module(lib).register
                libraries[django_lib] = self._parse_library(lib)
            except (InvalidTemplateLibrary, KeyError) as e:
                logger.error(f"Failed to parse django templatetag {django_lib}: {e}")
                continue

        for app_config in apps.get_app_configs():
            app = app_config.name
            try:
                templatetag_mod = __import__(app + ".templatetags", {}, {}, [""])
            except ImportError:
                continue

            try:
                mod_path = inspect.getabsfile(templatetag_mod)
            except TypeError as e:
                logger.error(f"Failed getting path for ({app}) templatetags: {e}")
                continue
            mod_files = os.listdir(os.path.dirname(mod_path))
            tag_files = [i[:-3] for i in mod_files if i.endswith(".py") and i[0] != "_"]

            for taglib in tag_files:
                try:
                    lib = get_installed_libraries()[taglib]
                    lib = importlib.import_module(lib).register
                except (InvalidTemplateLibrary, KeyError) as e:
                    logger.error(f"Failed to parse library ({taglib}): {e}")
                    continue

                libraries[taglib] = self._parse_library(lib)

        # Add node tags
        for lib_name, tags in LIBRARIES_NODE_TAGS.items():
            if lib_name in libraries:
                for tag, options in tags.items():
                    if tag in libraries[lib_name]["tags"]:
                        libraries[lib_name]["tags"][tag]["inner_tags"] = options.get(
                            "inner_tags", []
                        )
                        libraries[lib_name]["tags"][tag]["closing_tag"] = options.get(
                            "closing_tag"
                        )

        return libraries

    def _parse_library(self, lib) -> dict:
        return {
            "tags": {
                name: {
                    "docs": func.__doc__.strip() if func.__doc__ else "",
                    "source": self.get_source_from_type(func),
                }
                for name, func in lib.tags.items()
            },
            "filters": {
                name: {
                    "docs": func.__doc__.strip() if func.__doc__ else "",
                    "source": self.get_source_from_type(func),
                }
                for name, func in lib.filters.items()
            },
        }

    # Templates
    # ---------------------------------------------------------------------------------
    def get_templates(self):
        template_files = {}
        default_engine = Engine.get_default()
        for templates_dir in [
            *default_engine.dirs,
            *get_app_template_dirs("templates"),
        ]:
            for root, dirs, files in os.walk(templates_dir):
                for file in files:
                    template_name = os.path.relpath(
                        os.path.join(root, file), templates_dir
                    )

                    if template_name in template_files:
                        # Skip already procecesed template
                        # (template have duplicates because other apps can override)
                        continue

                    # Get used template (other apps can override templates)
                    template_files[template_name] = self._parse_template(
                        self._get_template(default_engine, template_name),
                    )
        return template_files

    def add_template_context_for_view(self, view):
        if not issubclass(view, View):
            # Ensure only class-based views (CBVs) are allowed; function-based
            # views (FBVs) are not supported
            return

        view_obj = view(request=MagicMock())

        try:
            template_name = view_obj.get_template_names()[0]
        except Exception:
            template_name = getattr(view, "template_name", None)

        if template_name in self.templates:
            if issubclass(view, SingleObjectMixin) and hasattr(view, "model"):
                context = {"object": self.get_type_full_name(view.model)}
                try:
                    context_name = view_obj.get_context_object_name(view.model)
                    if context_name:
                        context[context_name] = context["object"]
                except Exception:
                    pass
                self.templates[template_name]["context"].update(context)
            if issubclass(view, MultipleObjectMixin):
                try:
                    paginator = self.get_type_full_name(view.paginator_class)
                except Exception:
                    paginator = "django.core.paginator.Paginator"

                self.templates[template_name]["context"].update(
                    {
                        "paginator": paginator,
                        "page_obj": "django.core.paginator.Page",
                        "is_paginated": "bool",
                        "object_list": "django.db.models.QuerySet",
                    }
                )
            if issubclass(view, FormMixin) and hasattr(view, "form_class"):
                self.templates[template_name]["context"].update(
                    {"form": self.get_type_full_name(view.form_class)}
                )

    def _parse_template(self, template: Template) -> dict:
        extends = None
        blocks = set()
        for line in template.content.splitlines():
            if match := self.re_extends.match(line):
                extends = match.group(1)
            if match := self.re_block.match(line):
                blocks.add(match.group(1))

        path = ""
        if template.path.startswith(self.project_src_path):
            path = (
                f"src:{template.path.removeprefix(self.project_src_path).lstrip('/')}"
            )
        elif template.path.startswith(sys.prefix):
            path = f"env:{template.path.removeprefix(sys.prefix).lstrip('/')}"

        return {
            "path": path,
            "extends": extends,
            "blocks": list(blocks),
            "context": {},
        }

    def _get_template(self, engine: Engine, template_name: str) -> Template:
        for loader in engine.template_loaders:
            for origin in loader.get_template_sources(template_name):
                try:
                    return Template(
                        path=str(origin),
                        name=template_name,
                        content=loader.get_contents(origin),
                    )
                except Exception:
                    pass
        return Template(name=template_name)

    # Global context
    # ---------------------------------------------------------------------------------
    def get_global_template_context(self):
        global_context = {
            # builtins
            "True": None,
            "False": None,
            "None": None,
        }

        # Update object types
        TEMPLATE_CONTEXT_PROCESSORS["django.contrib.auth.context_processors.auth"][
            "user"
        ] = f"{get_user_model().__module__}.{get_user_model().__name__}"

        for context_processor in Engine.get_default().template_context_processors:
            module_path = ".".join(
                [context_processor.__module__, context_processor.__name__]
            )
            if context := TEMPLATE_CONTEXT_PROCESSORS.get(module_path):
                global_context.update(context)
        return global_context

    # Third party: Wagtail
    # ---------------------------------------------------------------------------------
    def collect_for_wagtail(self):
        try:
            from wagtail.models import Page
        except ImportError:
            return
        for model in apps.get_models():
            if issubclass(model, Page) and model.template in self.templates:
                self.templates[model.template]["context"].update(
                    {
                        "page": model.__module__ + "." + model.__name__,
                        "self": model.__module__ + "." + model.__name__,
                    }
                )
                if model.context_object_name:
                    self.templates[model.template]["context"][
                        model.context_object_name
                    ] = (model.__module__ + "." + model.__name__)


#######################################################################################
# CLI
#######################################################################################
def get_default_django_settings_module():
    try:
        # Patch django execute to prevent it from running when calling main
        with patch(
            "django.core.management.execute_from_command_line", return_value=None
        ):
            from manage import main

            # Need to run main becuase default env is set in main function
            main()

        return os.getenv("DJANGO_SETTINGS_MODULE")
    except ImportError:
        return ""


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="""
            Collect django project information for LSP.
            pure python script that is runned in project environment for using django.
        """
    )
    parser.add_argument("--django-settings-module", action="store", type=str)
    parser.add_argument("--project-src", action="store", type=str)
    args = parser.parse_args()

    project_src_path = args.project_src if args.project_src else os.getcwd()
    sys.path.insert(0, project_src_path)

    django_settings_module = (
        args.django_settings_module
        if args.django_settings_module
        else get_default_django_settings_module()
    )
    if args.django_settings_module:
        os.environ.setdefault(
            "DJANGO_SETTINGS_MODULE",
            django_settings_module,
        )

    # Enable error logging to stderr
    logging.basicConfig(
        level=logging.ERROR,
        format="%(message)s",
        handlers=[logging.StreamHandler(sys.stderr)],
    )

    django.setup()

    collector = DjangoIndexCollector(project_src_path)
    collector.collect()

    print(collector.to_json())



BUILTIN = "__builtins__"

FALLBACK_DJANGO_DATA = {
    "file_watcher_globs": ["**/templates/**", "**/templatetags/**", "**/static/**"],
    "static_files": [],
    "urls": {},
    "libraries": {
        "__builtins__": {
            "tags": {
                "autoescape": {"inner_tags": [], "closing_tag": "endautoescape"},
                "comment": {"inner_tags": [], "closing_tag": "endcomment"},
                "cycle": {},
                "csrf_token": {},
                "debug": {},
                "filter": {"inner_tags": [], "closing_tag": "endfilter"},
                "firstof": {},
                "for": {"inner_tags": ["empty"], "closing_tag": "endfor"},
                "if": {"inner_tags": ["else", "elif"], "closing_tag": "endif"},
                "ifchanged": {"inner_tags": [], "closing_tag": "endifchanged"},
                "load": {},
                "lorem": {},
                "now": {},
                "regroup": {},
                "resetcycle": {},
                "spaceless": {"inner_tags": [], "closing_tag": "endspaceless"},
                "templatetag": {},
                "url": {},
                "verbatim": {"inner_tags": [], "closing_tag": "endverbatim"},
                "widthratio": {},
                "with": {"inner_tags": [], "closing_tag": "endwith"},
                "block": {"inner_tags": [], "closing_tag": "endblock"},
                "extends": {},
                "include": {},
            },
            "filters": {
                "addslashes": {},
                "capfirst": {},
                "escapejs": {},
                "json_script": {},
                "floatformat": {},
                "iriencode": {},
                "linenumbers": {},
                "lower": {},
                "make_list": {},
                "slugify": {},
                "stringformat": {},
                "title": {},
                "truncatechars": {},
                "truncatechars_html": {},
                "truncatewords": {},
                "truncatewords_html": {},
                "upper": {},
                "urlencode": {},
                "urlize": {},
                "urlizetrunc": {},
                "wordcount": {},
                "wordwrap": {},
                "ljust": {},
                "rjust": {},
                "center": {},
                "cut": {},
                "escape": {},
                "force_escape": {},
                "linebreaks": {},
                "linebreaksbr": {},
                "safe": {},
                "safeseq": {},
                "striptags": {},
                "dictsort": {},
                "dictsortreversed": {},
                "first": {},
                "join": {},
                "last": {},
                "length": {},
                "length_is": {},
                "random": {},
                "slice": {},
                "unordered_list": {},
                "add": {},
                "get_digit": {},
                "date": {},
                "time": {},
                "timesince": {},
                "timeuntil": {},
                "default": {},
                "default_if_none": {},
                "divisibleby": {},
                "yesno": {},
                "filesizeformat": {},
                "pluralize": {},
                "phone2numeric": {},
                "pprint": {},
            },
        },
        "cache": {
            "tags": {"cache": {"inner_tags": [], "closing_tag": "endcache"}},
            "filters": {},
        },
        "i18n": {
            "tags": {
                "get_available_languages": {},
                "get_language_info": {},
                "get_language_info_list": {},
                "get_current_language": {},
                "get_current_language_bidi": {},
                "trans": {},
                "translate": {},
                "blocktrans": {},
                "blocktranslate": {},
                "language": {"inner_tags": [], "closing_tag": "endlanguage"},
            },
            "filters": {
                "language_name": {},
                "language_name_translated": {},
                "language_name_local": {},
                "language_bidi": {},
            },
        },
        "l10n": {
            "tags": {"localize": {"inner_tags": [], "closing_tag": "endlocalize"}},
            "filters": {"localize": {}, "unlocalize": {}},
        },
        "static": {
            "tags": {"get_static_prefix": {}, "get_media_prefix": {}, "static": {}},
            "filters": {},
        },
        "tz": {
            "tags": {
                "localtime": {"inner_tags": [], "closing_tag": "endlocaltime"},
                "timezone": {"inner_tags": [], "closing_tag": "endtimezone"},
                "get_current_timezone": {},
            },
            "filters": {"localtime": {}, "utc": {}, "timezone": {}},
        },
    },
    "templates": {},
    "global_template_context": {
        "csrf_token": {},
    },
}



import json
import logging
import os
import shutil
import subprocess
import uuid
from functools import cached_property

import jedi
from lsprotocol.types import (
    INITIALIZE,
    TEXT_DOCUMENT_COMPLETION,
    TEXT_DOCUMENT_DEFINITION,
    TEXT_DOCUMENT_HOVER,
    WORKSPACE_DID_CHANGE_WATCHED_FILES,
    CompletionList,
    CompletionOptions,
    CompletionParams,
    DefinitionParams,
    DidChangeWatchedFilesParams,
    DidChangeWatchedFilesRegistrationOptions,
    FileSystemWatcher,
    HoverParams,
    InitializeParams,
    Registration,
    RegistrationParams,
)
from pygls.server import LanguageServer

from djlsp import __version__
from djlsp.constants import FALLBACK_DJANGO_DATA
from djlsp.index import WorkspaceIndex
from djlsp.parser import TemplateParser

logger = logging.getLogger(__name__)


DJANGO_COLLECTOR_SCRIPT_PATH = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    "scripts",
    "django-collector.py",
)


class DjangoTemplateLanguageServer(LanguageServer):
    ENV_DIRECTORIES = [
        "env",
        ".env",
        "venv",
        ".venv",
    ]

    def __init__(self, *args):
        super().__init__(*args)
        self.file_watcher_id = str(uuid.uuid4())
        self.current_file_watcher_globs = []
        self.docker_compose_file = "docker-compose.yml"
        self.docker_compose_service = "django"
        self.django_settings_module = ""
        self.workspace_index = WorkspaceIndex()
        self.workspace_index.update(FALLBACK_DJANGO_DATA)
        self.jedi_project = jedi.Project(".")

    @cached_property
    def project_src_path(self):
        """Root path to src files, auto detect based on manage.py file"""
        for name in os.listdir(self.workspace.root_path):
            src_path = os.path.join(self.workspace.root_path, name)
            if os.path.exists(os.path.join(src_path, "manage.py")):
                return src_path
        return self.workspace.root_path

    @cached_property
    def project_env_path(self):
        for env_dir in self.ENV_DIRECTORIES:
            if os.path.exists(
                os.path.join(self.workspace.root_path, env_dir, "bin", "python")
            ):
                return os.path.join(self.workspace.root_path, env_dir)

    @property
    def docker_compose_path(self):
        return os.path.join(self.workspace.root_path, self.docker_compose_file)

    def set_initialization_options(self, options: dict):
        self.docker_compose_file = options.get(
            "docker_compose_file", self.docker_compose_file
        )
        self.docker_compose_service = options.get(
            "docker_compose_service", self.docker_compose_service
        )
        self.django_settings_module = options.get(
            "django_settings_module", self.django_settings_module
        )

    def set_file_watcher_capability(self):
        logger.info(
            f"Update file watcher patterns to: {self.current_file_watcher_globs}"
        )
        self.register_capability(
            RegistrationParams(
                registrations=[
                    Registration(
                        id=self.file_watcher_id,
                        method=WORKSPACE_DID_CHANGE_WATCHED_FILES,
                        register_options=DidChangeWatchedFilesRegistrationOptions(
                            watchers=[
                                FileSystemWatcher(glob_pattern=glob_pattern)
                                for glob_pattern in self.current_file_watcher_globs
                            ]
                        ),
                    )
                ]
            )
        )

    def get_django_data(self):
        self.workspace_index.src_path = self.project_src_path
        self.workspace_index.env_path = self.project_env_path
        self.jedi_project = jedi.Project(
            path=self.project_src_path, environment_path=self.project_env_path
        )

        if self.project_env_path:
            django_data = self._get_django_data_from_python_path(
                os.path.join(self.project_env_path, "bin", "python")
            )
        elif self._has_valid_docker_service():
            django_data = self._get_django_data_from_docker()
        elif python_path := shutil.which("python3"):
            # Try getting data with global python installtion
            django_data = self._get_django_data_from_python_path(python_path)
        else:
            django_data = None

        if django_data:
            # TODO: Maybe validate data
            self.workspace_index.update(django_data)
            logger.info("Collected project Django data:")
            logger.info(f" - Libraries: {len(django_data['libraries'])}")
            logger.info(f" - Templates: {len(django_data['templates'])}")
            logger.info(f" - Static files: {len(django_data['static_files'])}")
            logger.info(f" - Urls: {len(django_data['urls'])}")
            logger.info(
                f" - Global context: {len(django_data['global_template_context'])}"
            )
        else:
            logger.info("Could not collect Django data")

        if set(self.workspace_index.file_watcher_globs) != set(
            self.current_file_watcher_globs
        ):
            self.current_file_watcher_globs = self.workspace_index.file_watcher_globs
            self.set_file_watcher_capability()

    def _get_django_data_from_python_path(self, python_path):
        logger.info(f"Collection django data from local python path: {python_path}")

        command = list(
            filter(
                None,
                [
                    python_path,
                    DJANGO_COLLECTOR_SCRIPT_PATH,
                    (
                        f"--django-settings-module={self.django_settings_module}"  # noqa: E501
                        if self.django_settings_module
                        else None
                    ),
                    f"--project-src={self.project_src_path}",
                ],
            )
        )

        logger.debug(f"Collector command: {' '.join(command)}")

        try:
            return json.loads(subprocess.check_output(command).decode())
        except Exception as e:
            logger.error(e)
            return False

    def _has_valid_docker_service(self):
        if os.path.exists(self.docker_compose_path):
            services = (
                subprocess.check_output(
                    [
                        "docker",
                        "compose",
                        f"--file={self.docker_compose_path}",
                        "config",
                        "--services",
                    ]
                )
                .decode()
                .splitlines()
            )
            return self.docker_compose_service in services
        return False

    def _get_django_data_from_docker(self):
        logger.info(
            f"Collecting django data from docker {self.docker_compose_file}:{self.docker_compose_service}"  # noqa: E501
        )

        docker_image = self._get_docker_image()
        if not docker_image:
            return False

        docker_run_command = list(
            filter(
                None,
                [
                    "docker",
                    "run",
                    "--rm",
                    f"--volume={DJANGO_COLLECTOR_SCRIPT_PATH}:/django-collector.py",
                    f"--volume={self.project_src_path}:/src",
                    docker_image,
                    "python",
                    "/django-collector.py",
                    (
                        f"--django-settings-module={self.django_settings_module}"
                        if self.django_settings_module
                        else None
                    ),
                    "--project-src=/src",
                ],
            )
        )

        logger.debug(f"Collector command: {' '.join(docker_run_command)}")

        try:
            return json.loads(subprocess.check_output(docker_run_command).decode())
        except Exception as e:
            logger.error(e)
            return False

    def _get_docker_image(self):
        try:
            # Make sure image is created
            subprocess.check_call(
                [
                    "docker",
                    "compose",
                    f"--file={self.docker_compose_path}",
                    "create",
                    "--no-recreate",
                    self.docker_compose_service,
                ]
            )
        except Exception as e:
            logger.error(e)
            return None

        try:
            images = json.loads(
                subprocess.check_output(
                    [
                        "docker",
                        "compose",
                        f"--file={self.docker_compose_path}",
                        "images",
                        self.docker_compose_service,
                        "--format=json",
                    ]
                )
            )
        except Exception as e:
            logger.error(e)
            return None

        if images:
            return images[0]["ID"]
        return None


server = DjangoTemplateLanguageServer("django-template-lsp", __version__)


@server.feature(INITIALIZE)
def initialized(ls: DjangoTemplateLanguageServer, params: InitializeParams):
    logger.info(f"COMMAND: {INITIALIZE}")
    logger.debug(f"OPTIONS: {params.initialization_options}")
    if params.initialization_options:
        ls.set_initialization_options(params.initialization_options)
    ls.get_django_data()


@server.feature(
    TEXT_DOCUMENT_COMPLETION,
    CompletionOptions(trigger_characters=[" ", "|", "'", '"', "."]),
)
def completions(ls: DjangoTemplateLanguageServer, params: CompletionParams):
    logger.info(f"COMMAND: {TEXT_DOCUMENT_COMPLETION}")
    logger.debug(f"PARAMS: {params}")
    try:
        return CompletionList(
            is_incomplete=False,
            items=TemplateParser(
                workspace_index=ls.workspace_index,
                jedi_project=ls.jedi_project,
                document=server.workspace.get_document(params.text_document.uri),
            ).completions(params.position.line, params.position.character),
        )
    except Exception as e:
        logger.error(e)
        return CompletionList(items=[])


@server.feature(TEXT_DOCUMENT_HOVER)
def hover(ls: DjangoTemplateLanguageServer, params: HoverParams):
    logger.info(f"COMMAND: {TEXT_DOCUMENT_HOVER}")
    logger.debug(f"PARAMS: {params}")
    try:
        return TemplateParser(
            workspace_index=ls.workspace_index,
            jedi_project=ls.jedi_project,
            document=server.workspace.get_document(params.text_document.uri),
        ).hover(params.position.line, params.position.character)
    except Exception as e:
        logger.error(e)
        return None


@server.feature(TEXT_DOCUMENT_DEFINITION)
def goto_definition(ls: DjangoTemplateLanguageServer, params: DefinitionParams):
    logger.info(f"COMMAND: {TEXT_DOCUMENT_DEFINITION}")
    logger.debug(f"PARAMS: {params}")
    try:
        return TemplateParser(
            workspace_index=ls.workspace_index,
            jedi_project=ls.jedi_project,
            document=ls.workspace.get_document(params.text_document.uri),
        ).goto_definition(params.position.line, params.position.character)
    except Exception as e:
        logger.error(e)
        return None


@server.feature(WORKSPACE_DID_CHANGE_WATCHED_FILES)
def files_changed(
    ls: DjangoTemplateLanguageServer, params: DidChangeWatchedFilesParams
):
    logger.info(f"COMMAND: {WORKSPACE_DID_CHANGE_WATCHED_FILES}")
    logger.debug(f"PARAMS: {params}")
    # TODO: Do partial collect based on changed file type
    ls.get_django_data()
